#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, NoReturn, Optional, Tuple

from etl_entities import HWM, Column, ColumnHWM
from pydantic import Field, root_validator, validator
from typing_extensions import Protocol

from onetl.db.db_reader.db_reader import DBReader
from onetl.hwm import Statement
from onetl.hwm.store import HWMClassRegistry
from onetl.impl import FrozenModel
from onetl.strategy.batch_hwm_strategy import BatchHWMStrategy
from onetl.strategy.hwm_strategy import HWMStrategy
from onetl.strategy.strategy_manager import StrategyManager

log = getLogger(__name__)

if TYPE_CHECKING:
    from pyspark.sql.dataframe import DataFrame


# ColumnHWM has abstract method serialize_value, so it's not possible to create a class instance
# small hack to bypass this exception
class MockColumnHWM(ColumnHWM):
    def serialize_value(self):
        """Fake implementation of ColumnHWM.serialize_value"""


class StrategyHelper(Protocol):
    def save(self, df: DataFrame) -> DataFrame:
        """Saves HWM value to HWMStore"""

    def get_boundaries(self) -> tuple[Statement | None, Statement | None]:
        """Returns ``(min_boundary, max_boundary)`` for applying HWM to source"""


class NonHWMStrategyHelper(FrozenModel):
    reader: DBReader

    def get_boundaries(self) -> Tuple[Optional[Statement], Optional[Statement]]:
        return None, None

    @root_validator(pre=True)
    def validate_current_strategy(cls, values):
        reader = values.get("reader")
        strategy = StrategyManager.get_current()

        if isinstance(strategy, HWMStrategy):
            raise ValueError(
                f"{strategy.__class__.__name__} cannot be used "
                f"without `hwm_column` passed into {reader.__class__.__name__}",
            )

        return values

    def save(self, df: DataFrame) -> DataFrame:
        return df


class HWMStrategyHelper(FrozenModel):
    reader: DBReader
    hwm_column: Column
    hwm_expression: Optional[str] = None
    strategy: HWMStrategy = Field(default_factory=StrategyManager.get_current)

    class Config:
        validate_all = True

    @validator("strategy", always=True, pre=True)
    def validate_strategy_is_hwm(cls, strategy, values):
        reader = values.get("reader")

        if not isinstance(strategy, HWMStrategy):
            raise ValueError(
                f"{strategy.__class__.__name__} cannot be used "
                f"with `hwm_column` passed into {reader.__class__.__name__}",
            )

        return strategy

    @validator("strategy", always=True)
    def validate_strategy_matching_reader(cls, strategy, values):
        if strategy.hwm is None:
            return strategy

        reader = values.get("reader")
        hwm_column = values.get("hwm_column")

        if not isinstance(strategy.hwm, ColumnHWM):
            cls.raise_wrong_hwm_type(reader, type(strategy.hwm))

        if strategy.hwm.source != reader.source or strategy.hwm.column != hwm_column:
            raise ValueError(
                f"{reader.__class__.__name__} was created "
                f"with `hwm_column={reader.hwm_column}` and `source={reader.source}` "
                f"but current HWM is created for ",
                f"`column={strategy.hwm.column}` and `source={strategy.hwm.source}` ",
            )

        return strategy

    @validator("strategy", always=True)
    def init_hwm(cls, strategy, values):
        reader = values.get("reader")
        hwm_column = values.get("hwm_column")

        if strategy.hwm is None:
            # Small hack used only to generate qualified_name
            strategy.hwm = MockColumnHWM(source=reader.source, column=hwm_column)

        if not strategy.hwm:
            strategy.fetch_hwm()

        hwm_type: type[HWM] | None = type(strategy.hwm)
        if hwm_type == MockColumnHWM:
            # Remove HWM type set by hack above
            hwm_type = None

        detected_hwm_type = cls.detect_hwm_column_type(reader, hwm_column)

        if not hwm_type:
            hwm_type = detected_hwm_type

        if hwm_type != detected_hwm_type:
            raise TypeError(
                f'Type of "{hwm_column}" column is matching '
                f'"{detected_hwm_type.__name__}" which is different from "{hwm_type.__name__}"',
            )

        if hwm_type == MockColumnHWM or not issubclass(hwm_type, ColumnHWM):
            cls.raise_wrong_hwm_type(reader, hwm_type)

        strategy.hwm = hwm_type(source=reader.source, column=hwm_column, value=strategy.hwm.value)
        return strategy

    @validator("strategy", always=True)
    def detect_hwm_column_boundaries(cls, strategy, values):
        if not isinstance(strategy, BatchHWMStrategy):
            return strategy

        if strategy.has_upper_limit and (strategy.has_lower_limit or strategy.hwm):
            # values already set by previous reader runs within the strategy
            return strategy

        reader = values.get("reader")
        hwm_column = values.get("hwm_column")
        hwm_expression = values.get("hwm_expression")

        min_hwm_value, max_hwm_value = reader.get_min_max_bounds(hwm_column.name, hwm_expression)
        if min_hwm_value is None or max_hwm_value is None:
            raise ValueError(
                "Unable to determine max and min values. ",
                f"Table '{reader.source}' column '{hwm_column}' cannot be used as `hwm_column`",
            )

        if not strategy.has_lower_limit and not strategy.hwm:
            strategy.start = min_hwm_value

        if not strategy.has_upper_limit:
            strategy.stop = max_hwm_value

        return strategy

    @staticmethod
    def raise_wrong_hwm_type(reader: DBReader, hwm_type: type[HWM]) -> NoReturn:
        raise ValueError(
            f"{hwm_type.__name__} cannot be used with {reader.__class__.__name__}",
        )

    @staticmethod
    def detect_hwm_column_type(reader: DBReader, hwm_column: Column) -> type[HWM]:
        schema = {field.name.casefold(): field for field in reader.get_df_schema()}
        column = hwm_column.name.casefold()
        hwm_column_type = schema[column].dataType.typeName()
        return HWMClassRegistry.get(hwm_column_type)

    def save(self, df: DataFrame) -> DataFrame:
        from pyspark.sql import functions as F  # noqa: N812

        log.info("|DBReader| Calculating max value for column %r in the dataframe...", self.hwm_column.name)
        max_df = df.select(F.max(self.hwm_column.name).alias("max_value"))
        row = max_df.collect()[0]
        max_hwm_value = row["max_value"]
        log.info("|DBReader| Max value is: %r", max_hwm_value)

        self.strategy.update_hwm(max_hwm_value)
        return df

    def get_boundaries(self) -> tuple[Statement | None, Statement | None]:
        start_from: Statement | None = None
        end_at: Statement | None = None
        hwm: ColumnHWM | None = self.strategy.hwm  # type: ignore

        if hwm is None:
            return None, None

        if self.strategy.current_value is not None:
            start_from = Statement(
                expression=self.hwm_expression or hwm.name,
                operator=self.strategy.current_value_comparator,
                value=self.strategy.current_value,
            )

        if self.strategy.next_value is not None:
            end_at = Statement(
                expression=self.hwm_expression or hwm.name,
                operator=self.strategy.next_value_comparator,
                value=self.strategy.next_value,
            )

        return start_from, end_at



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from onetl.hwm.store.base_hwm_store import BaseHWMStore
from onetl.hwm.store.hwm_class_registry import HWMClassRegistry, register_hwm_class
from onetl.hwm.store.hwm_store_class_registry import (
    HWMStoreClassRegistry,
    default_hwm_store_class,
    detect_hwm_store,
    register_hwm_store_class,
)
from onetl.hwm.store.hwm_store_manager import HWMStoreManager
from onetl.hwm.store.memory_hwm_store import MemoryHWMStore
from onetl.hwm.store.yaml_hwm_store import YAMLHWMStore



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, TypeVar

from onetl.base import BaseReadableFileFormat, BaseWritableFileFormat
from onetl.hooks import slot, support_hooks
from onetl.impl import GenericOptions

if TYPE_CHECKING:
    from pyspark.sql import DataFrameReader, DataFrameWriter

T = TypeVar("T")

PROHIBITED_OPTIONS = frozenset(
    (
        "ignoreCorruptFiles",
        "ignoreMissingFiles",
        "modifiedAfter",
        "modifiedBefore",
        "pathGlobFilter",
        "recursiveFileLookup",
    ),
)


@support_hooks
class ReadOnlyFileFormat(BaseReadableFileFormat, GenericOptions):
    name: ClassVar[str]

    class Config:
        prohibited_options = PROHIBITED_OPTIONS

    @slot
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader:
        options = self.dict(by_alias=True)
        return reader.format(self.name).options(**options)


@support_hooks
class WriteOnlyFileFormat(BaseWritableFileFormat, GenericOptions):
    name: ClassVar[str]

    class Config:
        prohibited_options = PROHIBITED_OPTIONS

    @slot
    def apply_to_writer(self, writer: DataFrameWriter) -> DataFrameWriter:
        options = self.dict(by_alias=True)
        return writer.format(self.name).options(**options)


class ReadWriteFileFormat(ReadOnlyFileFormat, WriteOnlyFileFormat):
    pass



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from onetl._util.version import Version


def get_default_scala_version(spark_version: Version) -> Version:
    """
    Get default Scala version for specific Spark version
    """
    if spark_version.major < 3:
        return Version(2, 11)
    return Version(2, 12)



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import logging
import operator
import os
from typing import Any, Callable, Collection, Optional

from etl_entities import HWM

from onetl.hwm.store import HWMStoreManager
from onetl.log import log_collection, log_with_indent
from onetl.strategy.base_strategy import BaseStrategy

log = logging.getLogger(__name__)


class HWMStrategy(BaseStrategy):
    hwm: Optional[HWM] = None

    @property
    def current_value(self) -> Any:
        if self.hwm:
            return self.hwm.value

        return super().current_value

    @property
    def current_value_comparator(self) -> Callable:
        return operator.gt

    @property
    def next_value_comparator(self) -> Callable:
        return operator.le

    def update_hwm(self, value: Any) -> None:
        if self.hwm is not None and value is not None:
            self.hwm.update(value)

    def enter_hook(self) -> None:
        # if HWM is already set (during previous run),
        # try to fetch its value using qualified_name
        self.fetch_hwm()

    def fetch_hwm(self) -> None:
        class_name = self.__class__.__name__

        if self.hwm is not None:
            hwm_store = HWMStoreManager.get_current()

            log.info("|%s| Loading HWM from %s:", class_name, hwm_store.__class__.__name__)
            log_with_indent(log, "qualified_name = %r", self.hwm.qualified_name)

            result = hwm_store.get(self.hwm.qualified_name)

            if result is not None:
                self.hwm = result
                log.info("|%s| Got HWM:", class_name)
                self._log_hwm(self.hwm)
            else:
                log.warning(
                    "|%s| HWM does not exist in %r. ALL ROWS/FILES WILL BE READ!",
                    class_name,
                    hwm_store.__class__.__name__,
                )
        else:
            # entering strategy context, HWM will be set later by DBReader.run or FileDownloader.run
            log.debug("|%s| HWM will not be loaded, skipping", class_name)

    def exit_hook(self, failed: bool = False) -> None:
        if not failed:
            self.save_hwm()

    def save_hwm(self) -> None:
        class_name = self.__class__.__name__

        if self.hwm is not None:
            hwm_store = HWMStoreManager.get_current()

            log.info("|%s| Saving HWM to %r:", class_name, hwm_store.__class__.__name__)
            self._log_hwm(self.hwm)
            log_with_indent(log, "qualified_name = %r", self.hwm.qualified_name)

            location = hwm_store.save(self.hwm)
            log.info("|%s| HWM has been saved", class_name)

            if location:
                if isinstance(location, os.PathLike):
                    log_with_indent(log, "location = '%s'", os.fspath(location))
                else:
                    log_with_indent(log, "location = %r", location)
        else:
            log.debug("|%s| HWM value is not set, do not save", class_name)

    def _log_hwm(self, hwm: HWM) -> None:
        log_with_indent(log, "type = %s", hwm.__class__.__name__)

        if isinstance(hwm.value, Collection):
            log_collection(log, "value", hwm.value, max_items=10)
        else:
            log_with_indent(log, "value = %r", hwm.value)

    @classmethod
    def _log_exclude_fields(cls) -> set[str]:
        return super()._log_exclude_fields() | {"hwm"}



from onetl.connection.db_connection.dialect_mixins.support_columns_list import (
    SupportColumnsList,
)
from onetl.connection.db_connection.dialect_mixins.support_columns_none import (
    SupportColumnsNone,
)
from onetl.connection.db_connection.dialect_mixins.support_df_schema_none import (
    SupportDfSchemaNone,
)
from onetl.connection.db_connection.dialect_mixins.support_df_schema_struct import (
    SupportDfSchemaStruct,
)
from onetl.connection.db_connection.dialect_mixins.support_hint_none import (
    SupportHintNone,
)
from onetl.connection.db_connection.dialect_mixins.support_hint_str import (
    SupportHintStr,
)
from onetl.connection.db_connection.dialect_mixins.support_hwm_column_str import (
    SupportHWMColumnStr,
)
from onetl.connection.db_connection.dialect_mixins.support_hwm_expression_none import (
    SupportHWMExpressionNone,
)
from onetl.connection.db_connection.dialect_mixins.support_hwm_expression_str import (
    SupportHWMExpressionStr,
)
from onetl.connection.db_connection.dialect_mixins.support_name_any import (
    SupportNameAny,
)
from onetl.connection.db_connection.dialect_mixins.support_name_with_schema_only import (
    SupportNameWithSchemaOnly,
)
from onetl.connection.db_connection.dialect_mixins.support_where_none import (
    SupportWhereNone,
)
from onetl.connection.db_connection.dialect_mixins.support_where_str import (
    SupportWhereStr,
)



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# TODO: remove in 1.0.0

import textwrap
import warnings
from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from onetl.hwm.store import (
        BaseHWMStore,
        HWMClassRegistry,
        HWMStoreClassRegistry,
        HWMStoreManager,
        MemoryHWMStore,
        YAMLHWMStore,
        default_hwm_store_class,
        detect_hwm_store,
        register_hwm_class,
        register_hwm_store_class,
    )

__all__ = [  # noqa: WPS410
    "BaseHWMStore",
    "HWMClassRegistry",
    "register_hwm_class",
    "HWMStoreClassRegistry",
    "default_hwm_store_class",
    "detect_hwm_store",
    "register_hwm_store_class",
    "HWMStoreManager",
    "MemoryHWMStore",
    "YAMLHWMStore",
]


def __getattr__(name: str):
    if name not in __all__:
        raise ImportError(f"cannot import name {name!r} from {__name__!r}")

    message = f"""
        Imports from module {__name__!r} are deprecated since v0.6.0 and will be removed in v1.0.0.
        Please use instead:

        from onetl.hwm.store import {name}
    """

    warnings.warn(textwrap.dedent(message).strip(), category=UserWarning, stacklevel=2)
    return getattr(import_module("onetl.hwm.store"), name)



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py4j.java_gateway import JavaGateway
    from pyspark.sql import SparkSession


def get_java_gateway(spark_session: SparkSession) -> JavaGateway:
    """
    Get py4j Java gateway object
    """
    return spark_session._sc._gateway  # noqa: WPS437  # type: ignore


def try_import_java_class(spark_session: SparkSession, name: str):
    """
    Try to import Java class by name
    """
    gateway = get_java_gateway(spark_session)
    klass = getattr(gateway.jvm, name)
    gateway.help(klass, display=False)
    return klass



#  Copyright 2023 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import ClassVar

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    _forward_refs_updated: ClassVar[bool] = False

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        extra = "forbid"
        underscore_attrs_are_private = True

    def __init__(self, **kwargs):
        if not self._forward_refs_updated:
            # if pydantic cannot detect type hints (referenced class is not imported yet),
            # it wraps annotation with ForwardRef(...), which should be resolved before creating the instance.
            # so using a small hack to detect all those refs and update them
            # when first object instance is being created
            refs = self._forward_refs()
            self.__class__.update_forward_refs(**refs)
            self.__class__._forward_refs_updated = True  # noqa: WPS437
        super().__init__(**kwargs)

    @classmethod
    def _forward_refs(cls) -> dict[str, type]:
        refs: dict[str, type] = {}
        for item in dir(cls):
            if item.startswith("_") or item.startswith("package"):
                continue

            value = getattr(cls, item)
            if isinstance(value, type):
                refs[item] = value

        return refs
